/* Calendar 0.1/REXX - Small textmode calendar
 * by Anton Dushko (2:5030/611.9)
 * This stuff is free. Comments & suggestions are welcome.
 */

/* Startup */
Trace Off
Address CMD
Signal On Halt Name CtrlBreak
Signal On Error Name IntError
Call RxFuncAdd "SysLoadFuncs", "RexxUtil", "SysLoadFuncs"
Call SysLoadFuncs

/* Read configuration */
Call ConfigRead
/* Load templates */
Call TemplateLoad("Calendar.TPL")
/* Initialize day names */
Do I = 0 To 6
  Calendar.I = Word(INIVars.DayNames, I + 1)
End I

/* Current date */
Calendar.Date = Date("S")
Calendar.CurrentYear = Format(Left(Calendar.Date, 4))
Calendar.CurrentMonth = Format(SubStr(Calendar.Date, 5, 2))
Calendar.CurrentDay = Format(Right(Calendar.Date, 2))

/* Display calendar */
Call SysCurState "OFF"
Call CharOut "CON", TemplateParse(Template.Calendar)

/* Starting values */
DisplayedMonth = Calendar.CurrentMonth
DisplayedYear = Calendar.CurrentYear
BufferRegenFrom = 1
BufferRegenTo = 3

/* Main loop */
Do Forever

  /* Buffers regeneration */
  Do I = BufferRegenFrom To BufferRegenTo
    Parse Value DateMYFix(DisplayedMonth + I - 2, DisplayedYear) With Calendar.Month","Calendar.Year .
    Calendar.MonthName = Word(INIVars.MonthNames, Calendar.Month)
    If Calendar.Month = Calendar.CurrentMonth &,
       Calendar.Year = Calendar.CurrentYear Then
      Calendar.Day = Calendar.CurrentDay
    Else
      Calendar.Day = ""
    Call CalendarFillMonthGrid
    Buffer.I = TemplateParse(Template.Month)
  End I

  /* Years regenaration */
  Parse Value DateMYFix(DisplayedMonth - 1, DisplayedYear) With ","Calendar.Year1 .
  Parse Value DateMYFix(DisplayedMonth + 1, DisplayedYear) With ","Calendar.Year2 .

  /* Display calendar */
  Call CharOut "CON", TemplateParse(Template.Years) ||,
                      Buffer.1 || Template.MonthShift ||,
                      Buffer.2 || Template.MonthShift ||,
                      Buffer.3 || Template.MonthShiftBack

  /* Blocking input*/
  Char = SysGetKey("NOECHO")
  If Char = "E0"x Then
    Char = Char || (SysGetKey("NOECHO"))

  /* User interaction*/
  Select
    /* Exit */
    When Char = "1B"x Then
      Leave
    /* Current month */
    When Char = "E047"x Then Do
      DisplayedMonth = Calendar.CurrentMonth
      DisplayedYear = Calendar.CurrentYear
      BufferRegenFrom = 1
      BufferRegenTo = 3
    End
    /* Previous month */
    When Char = "E049"x Then Do
      Parse Value DateMYFix(DisplayedMonth - 1, DisplayedYear) With DisplayedMonth","DisplayedYear .
      BufferRegenFrom = 1
      BufferRegenTo = 1
      Buffer.3 = Buffer.2
      Buffer.2 = Buffer.1
    End
    /* Next month */
    When Char = "E051"x Then Do
      Parse Value DateMYFix(DisplayedMonth + 1, DisplayedYear) With DisplayedMonth","DisplayedYear .
      BufferRegenFrom = 3
      BufferRegenTo = 3
      Buffer.1 = Buffer.2
      Buffer.2 = Buffer.3
    End
    Otherwise
      BufferRegenFrom = 0
      BufferRegenTo = 0
  End
End

/* Close calendar */
Call CharOut "CON", Template.CalendarEnd
Call SysCurState "ON"

/* All OK */
Exit 0


/* Read & process initialization file */
ConfigRead: Procedure Expose INIVars.
  /* Get location of script */
  Parse Source . . SrcName
  ScriptPath = FileSpec("D", SrcName) || FileSpec("P", SrcName)
  ScriptName = Left(FileSpec("N", SrcName), LastPos(".", FileSpec("N", SrcName)) - 1)

  /* Read & process <Caller>.INI */
  Signal On Syntax Name INIError
  INILine = 0
  INIFile = ScriptPath || ScriptName || ".INI"
  Call Stream INIFile, "C", "OPEN READ"
  Do While Stream(INIFile) = "READY"
    INILine = INILine + 1
    Parse Value Translate(LineIn(INIFile), " ", "9"x) With Variable Value
    If DataType(Variable, "S") = 1 Then
      Interpret "INIVars." || Strip(Variable) || "=" || Strip(Value)
  End
  Call Stream INIFile, "C", "CLOSE"
  Signal On Syntax Name IntError
Return

/* Initialization file loader's error handler */
INIError:
  Say "INI error in" INIFile ", line" INILine
Exit -10


/* Calendar functions */

/* Fill calendar grid for one month */
CalendarFillMonthGrid: Procedure Expose Calendar. INIVars. Template.

  /* Calculate month's first day weekday & month length */
  MonthBaseDate = DateDMY2BaseDate(1, Calendar.Month, Calendar.Year) + INIVars.WeekStyle
  MonthBeginWeekDay = MonthBaseDate // 7
  MonthLength = DateMonthLength(Calendar.Month, Calendar.Year)
  Week = 1

  /* Fill leading space */
  Do WeekDay = 0 To MonthBeginWeekDay
    Calendar.WeekDay.Week = CalendarDayHilighting(WeekDay, "")
  End WeekDay

  /* Fill existing days */
  Do MonthDay = 1 To MonthLength
    WeekDay = (MonthBeginWeekDay + MonthDay - 1) // 7
    Calendar.WeekDay.Week = CalendarDayHilighting(WeekDay, MonthDay)
    If WeekDay = 6 Then 
      Week = Week + 1
  End MonthDay

  /* Fill trailing space */
  WeekDay = (MonthBeginWeekDay + MonthDay - 1) // 7
  Do Week = Week To 6
    Do WeekDay = WeekDay To 6
      Calendar.WeekDay.Week = CalendarDayHilighting(WeekDay, "")
    End WeekDay
    WeekDay = 0
  End Week
Return

/* Do the day hilighting */
CalendarDayHilighting: Procedure Expose Calendar. INIVars. Template.
  Parse Arg WeekDay, Calendar.Num
  /* Holiday */
  If Symbol("INIVars.Hilight."Calendar.Month"."Calendar.Num) = "VAR" Then
    Calendar.Attr = Value("INIVars.Hilight."Calendar.Month"."Calendar.Num)
  Else
    /* Weekend */
    If WeekDay >= INIVars.WeekendBeginDay Then
      Calendar.Attr = INIVars.Hilight.Weekend
    Else
      Calendar.Attr = INIVars.Hilight.Normal

  /* Current or any other day */
  If (Calendar.Num \= "") & (Calendar.Num = Calendar.Day) Then
  Do
    Calendar.Attr = Word(Calendar.Attr, Words(Calendar.Attr))
    Return TemplateParse(Template.DayCurrent)
  End
  Else
  Do
    Calendar.Attr = Word(Calendar.Attr, 1)
    Return TemplateParse(Template.DayNormal)
  End
Return 


/* Date functions */

/* Convert 2-digits year to full year */
DateYY2YYYY Procedure
  Arg Year
  /* No century given - add current */
  If Year < 100 Then
    /* Check for century crossing */
    If SubStr(Date("S"), 3, 2) < Year Then
      Return Left(Date("S"), 2) || Year - 100
    Else
      Return Left(Date("S"), 2) || Year
Return Year

/* Fix year boundary crossing */
DateMYFix: Procedure
  Parse Arg Month, Year
  Select
    /* Next year */
    When Month < 1 Then
      Return Month + 12 || "," || Year - 1
    /* Previous year */
    When Month > 12 Then
      Return Month - 12 || "," || Year + 1
    /* This year */
    Otherwise
      Return Month || "," || Year
  End
Return

/* Convert day-month-year date to basedate */
DateDMY2BaseDate: Procedure
  Parse Arg Day, Month, Year
Return 365 * (Year - 1) + (Year - 1) % 4 - (Year - 1) % 100 + (Year - 1) % 400,
       + DateMonthShift(Month, Year) + Day - 1

/* Number of days from begining of year till the start of month */
DateMonthShift: Procedure
  Arg Month, Year
  Select
    /* Jan - Feb */
    When Month < 3 Then
      Return (Month - 1) * 31
    /* Mar - Jul */
    When Month < 8 Then
      Return (Month - 1) * 30 + Month % 2 - 2 + DateIsLeapYear(Year)
    /* Aug - Dec */
    Otherwise
      Return (Month - 1) * 30 + (Month - 7) % 2 + 2 + DateIsLeapYear(Year)
   End
Return

/* Determine if year is leap year */
DateIsLeapYear: Procedure
  Arg Year
Return (Year // 4 = 0) - (Year // 100 = 0) + (Year // 400 = 0)

/* Length of month */
DateMonthLength: Procedure
  Arg Month, Year
  Select
    /* Jan - Feb */
    When Month = 2 Then
      Return 28 + DateIsLeapYear(Year)
    /* Mar - Jul */
    When Month < 8 Then
      Return 30 + Month // 2
    /* Aug - Dec */
    Otherwise
      Return 30 + (Month - 1) // 2
   End
Return


/* Templates */

/* Load a template set */
TemplateLoad: Procedure Expose Template.
  Parse Arg TPLFile

  /* Open TPL file */
  RC = Stream(TPLFile, "C", "OPEN READ")
  If RC <> "READY:" Then
    Signal TPLError
    
  /* Load & close TPL file */
  AllTemplates = CharIn(TPLFile,, Stream(TPLFile, "C", "QUERY SIZE"))
  Call Stream TPLFile, "C", "CLOSE"
  
  /* Process it */
  Signal On Syntax Name TPLError
  M = 0
  Do Forever
    /* Find begining of template in file */
    N = Pos("0D0A"x || "%+", AllTemplates, M + 1)
    /* Exit if no templates left */
    If N = 0 Then
      Leave
    /* Find end of template */
    M = Pos("%-%" || "0D0A"x, AllTemplates, N + 1)
    If M = 0 Then
      M = Length(AllTemplates)
    /* Find end of template name */
    K = Pos("%" || "0D0A"x, AllTemplates, N + 1)

    /* Gather template body & name */
    Interpret("Template." || SubStr(AllTemplates, N + 4, K - N - 4) || "= SubStr(AllTemplates, K + 3, M - K - 3)") 
  End
  Signal On Syntax Name IntError
Return 

/* Parse a template */
TemplateParse: Procedure Expose Calendar.
  Parse Arg Template
  
  Result = ""
  Do I = 1
    /* Find position of opening escape character */
    J = Pos("%", Template, I)
    /* Set it's position after the end of template if not found */
    If J = 0 Then
      J = Length(Template) + 1

    /* Add text before escape char to result */
    Result = Result || SubStr(Template, I, J - I)

    /* Move begining-of-text pointer */
    I = J + 1
    /* Leave if it falls out of string */
    If I > Length(Template) Then
      Leave I

    /* Find position of closing escape charater */
    J = Pos("%", Template, I)
    /* Leave if it falls out of string */
    If J = 0 Then 
      Leave I

    /* Add escape character to result if macro is empty */
    If I = J Then
    Do
      Result = Result || "%"
      Iterate I
    End

    /* Parse content of macro */
    Parse Value SubStr(Template, I, J - I) With Var","Len","Align","Pad .

    /* Default to left align & space padding */
    If Align = "" Then
      Align = "L"
    If Pad = "" Then
      Pad = " "

    /* Add macro value to result */
    Var = Translate(Var)
    If Symbol("Calendar." || Var) = "VAR" Then
      /* Select aligning */
      Select
        /* No aligning */
        When Len = "" Then
          Result = Result || Calendar.Var
        /* Left aligned */
        When Align = "L" Then
          Result = Result || Left(Calendar.Var, Len, Pad)
        /* Right aligned */
        When Align = "R" Then
          Result = Result || Right(Calendar.Var, Len, Pad)
        /* Centred */
        When Align = "C" Then
          Result = Result || Center(Calendar.Var, Len, Pad)
      End

    /* Move begining-of-text pointer */
    I = J
  End I
Return Result

/* TemplateDrop - drops all templates */
TemplateDrop: Procedure Expose Template.
  Drop Template.
Return

/* Template loader's error handler */
TPLError:
  Say "TPL error in" TPLFile "-" RC
Exit -20

/* Signal handlers */

IntError:
  Say "Internal error" RC "at line" SIGL
Exit -5

CtrlBreak:
  Say "User break"
Exit -1
